/*
 * Copyright (c) 2007, 2010 by Intevation GmbH
 *
 * Author: Bernhard Herzog <bh@intevation.de>
 *
 * This program is free software under the LGPL (>=v2.1)
 * Read the file LGPL.txt coming with the software for details.
 */

package de.intevation.osaas;

import java.io.BufferedWriter;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.UnsupportedEncodingException;
import java.io.Writer;
import java.net.HttpURLConnection;
import java.net.URL;
import java.net.URLEncoder;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.TimeZone;

public class OSAASClient {

    private URL osaasURL;
    private String basicauth;
    private Writer failLog;

    public OSAASClient(URL url, String username, String password,
                       File failLogFile) throws Exception {
        osaasURL = url;
        basicauth = "Basic " + base64encode(username + ":" + password);

        OutputStream out = new FileOutputStream(failLogFile, true);
        failLog = new BufferedWriter(new OutputStreamWriter(out, "UTF-8"));
    }

    private static String base64encode(String s) {
        try {
            ByteArrayOutputStream bytes = new ByteArrayOutputStream();
            OutputStream encoder = new Base64Encoder.OutputStream(bytes);
            encoder.write(s.getBytes("UTF-8"));
            encoder.close();
            String encoded = bytes.toString("US-ASCII");
            if (encoded.endsWith("\n")) {
                encoded = encoded.substring(0, encoded.length() - 1);
            }
            return encoded;
        }
        catch (Exception e) {
            e.printStackTrace();
        }
        return null;
    }

    private String formatTime(Calendar datetime) {
        return String.format("%1$tFT%1$tT.%1$tLZ", datetime);
    }

    private String getTimestamp() {
        return formatTime(new GregorianCalendar(TimeZone.getTimeZone("GMT")));
    }

    private void logFailedRequest(String request) {
        try {
            failLog.write(request);
            failLog.write("\n");
            failLog.flush();
        }
        catch (IOException e) {
            e.printStackTrace();
        }
    }

    private String buildRequest(String user, String queryString,
                                String wmsidextern, String wmsidintern,
                                Calendar startTime) {
        String[][] fields = {
            {"user", user},
            {"endtime", getTimestamp()},
            {"starttime", formatTime(startTime)},
            {"wmsidextern", wmsidextern},
            {"wmsidintern", wmsidintern},
            {"requeststring", queryString},
        };
        StringBuilder buf = new StringBuilder();

        for (int i = 0; i < fields.length; i++) {
            if (i > 0)
                buf.append("&");
            buf.append(fields[i][0] + "=");
            try {
                buf.append(URLEncoder.encode(fields[i][1], "UTF-8"));
            }
            catch (UnsupportedEncodingException e) {
                // We shouldn't get here because UTF-8 is one of the
                // standard encodings
                e.printStackTrace();
            }
        }

        return buf.toString();
    }

    public void sendRequest(String user, String queryString, String wmsidextern,
                            String wmsidintern, Calendar startTime) {
        String request = buildRequest(user, queryString, wmsidextern,
                                      wmsidintern, startTime);
        int responseCode = 0;
        try {
            HttpURLConnection conn =
                (HttpURLConnection)osaasURL.openConnection();
            conn.setDoOutput(true);
            conn.setRequestProperty("Authorization", basicauth);

            OutputStreamWriter wr =
                new OutputStreamWriter(conn.getOutputStream());


            wr.write(request);
            wr.close();

            responseCode = conn.getResponseCode();
        } catch (Exception e) {
            System.out.println("Could not send OSAAS request: " + e);
        }
        finally {
            if (responseCode < 200 || responseCode >= 300) {
                logFailedRequest(request);
            }
        }
    }
}
