#! /usr/bin/env python
#
# Copyright (C) 2010 by Intevation GmbH
# Authors:
# Ingo Weinzierl <ingo.weinzierl@intevation.de>
#
# This program is free software under the GPL (>=v2)
# Read the file GPL.txt coming with the software for details.
#

import logging
import psycopg2 as db
import sys
import urllib

from ConfigParser import SafeConfigParser
from optparse import OptionParser
from lxml import etree
from StringIO import StringIO
from urlparse import urlparse

DEFAULT_CONFIG  = "billing.cfg"
DEFAULT_LOGFILE = "billing.log"

REQUEST_POST = 1
REQUEST_GET  = 2

NS_WFS = "http://www.opengis.net/wfs"

SQL_SELECT_OPEN_BILLINGS = """SELECT id, wfsidintern, wfsidextern,
requestformat, rawrequest AS TEXT, numfeatures, billtime FROM requests
WHERE billtime IS NULL"""
SQL_UPDATE_NUM_FEATURES = """UPDATE requests SET numfeatures = %(num)s,
billtime = CURRENT_TIMESTAMP where id =
%(id)s"""

config=None

def read_config(filename):
    """Reads the configuration from the file given by filename."""
    config = SafeConfigParser()
    config.read([filename])
    return config

def get_config_option(section, option, default=None):
    if not config:
        print "No config given yet."
        return default

    if not config.has_option(section, option):
        return default
    return config.get(section, option)

def setup_logging():
    """Sets up the logging."""

    loglevel = logging.WARN
    level    = get_config_option("logging", "loglevel", 1)
    if level and int(level) >= 3:
        print "Setup logging to DEBUG"
        loglevel = logging.DEBUG
    elif level and int(level) == 2:
        print "Setup logging to INFO"
        loglevel = logging.INFO

    logfile = get_config_option("logging", "logfile", DEFAULT_LOGFILE)

    file_handler = logging.FileHandler(logfile)
    file_handler.setFormatter(logging.Formatter(
        "%(asctime)s %(levelname)s %(message)s"))
    file_handler.setLevel(loglevel)

    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(logging.Formatter(
        "--> %(asctime)s %(levelname)s %(message)s"))
    console_handler.setLevel(logging.WARN)

    global logger
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)
    logger.addHandler(console_handler)
    logger.addHandler(file_handler)


class Request(object):
    """This class represents a very simple request object.

    It stores some basic things like the server url, the request format
    (1=HTTP-POST, 2=HTTP-GET), its data, the number of features that have been
    retrieved by this request earlier and the time where the billing has taken
    place."""
    def __init__(self, id, wfsidintern, wfsidextern, requestformat,
                 rawrequest, numfeatures, cashtime):
        self.id = id
        self.wfsidintern   = wfsidintern
        self.wfsidextern   = wfsidextern
        self.requestformat = requestformat
        self.rawrequest    = rawrequest
        self.numfeatures   = numfeatures
        self.cashtime      = cashtime

    def handle_response(self, doc):
        """ This method handles the response object of the request. Currently,
        it searches for an xml element with the xpath
        '/FeatureCollection/@numberOfFeatures' where the number of features
        retrieved by this request should be stored. The default namespace for
        this xpath is NS_WFS. If a valid quantity is found by the xpath
        expression, the number of features is saved to database."""
        try:
            logger.debug("Response finished.")
            response = doc.xpath('/g:FeatureCollection/@numberOfFeatures',
                                 namespaces={'g': NS_WFS})
            if not response:
                logger.warn("GetFeature request retrieved no number of "
                            "features.")
                return False

            numFeatures = int(response[0])
            logger.info("GetFeature retrieved %d features." %
                        numFeatures)
            self.numfeatures = numFeatures
            return self.update_num_features()
        except ValueError, e:
            logger.error("The response did not retrieve a valid quantity.")
        return False

    def update_num_features(self):
        """ This method opens a database connection and updates the request
        object's number of features using the SQL query defined by
        SQL_UPDATE_NUM_FEATURES. The return value of this function is the number
        of features that have been written into database, or False if the update
        process failed."""
        db_conn = open_connection()

        if not db_conn:
            return False

        try:
            logger.debug("Execute update statement for request id %d" %
                         self.id)
            cur = db_conn.cursor()
            cur.execute(SQL_UPDATE_NUM_FEATURES, { 'id': self.id,
                                                   'num': self.numfeatures})
            db_conn.commit()
            close_connection(cur, db_conn)
            logger.info("Update for request id %d completed (%d features)"
                         % (self.id, self.numfeatures))
            return self.numfeatures
        except:
            logger.error("Could not update request id %r" % self.id)
            return False


class GETRequest(Request):
    """This request object represents HTTP-GET request objects and inherits from
    Request."""

    def is_get_feature_request(self):
        """This method returns True, if this request object represents an OGC
        WFS GetFeature request, otherwise it returns False."""
        if 'request=getfeature' in str(self.rawrequest).lower():
            return True
        return False

    def create_request_url(self):
        """This method modifies the URL of this GetFeature request and returns
        an URL which response is the number of features of the GetFeature
        request represented by this object."""
        req_url = None
        if "?" in self.wfsidintern:
            req_url = "%s&%s" % (self.wfsidintern, self.rawrequest)
        else:
            req_url = "%s?%s" % (self.wfsidintern, self.rawrequest)

        req_url = urllib.unquote(req_url)
        if 'resulttype' in req_url.lower():
            url_parts = urlparse(req_url)
            req_url   = req_url.split("?")[0]
            if url_parts and url_parts[4]:
                params = url_parts[4].split("&") or []
                for param in params:
                    if "?" in req_url:
                        req_url += "&"
                    else:
                        req_url += "?"

                    if 'resulttype' in param.lower():
                        param = "resultType=hits"
                    req_url += param
        else:
            req_url += '&resultType=hits'
        logger.debug("request url = %s" % req_url)
        return req_url

    def bill(self):
        """ This method should be called to take account for the request. If
        this object represents a GetFeature request, a modified version of this
        request is used to query the number of feature that would have been
        retrieved by this GetFeature request. If the modified version of this
        request returns a valid quantity, this number is used to update the row
        in the database. If everything was ok in this process, this method
        returns the number of features of this request, otherwise False."""
        req_url = self.create_request_url()
        try:
            doc = etree.parse(req_url)
            return self.handle_response(doc)
        except IOError:
            logger.error("Could not connect to server %s" %
                         self.wfsidintern)
        except:
            logger.error("Could not bill request with database id %r" %
                         self.id)
        return False


class POSTRequest(Request):
    """This request object represents HTTP-POST request objects and inherits
    from Request."""

    def is_get_feature_request(self):
        """This method returns True, if this request object represents an OGC
        WFS GetFeature request, otherwise it returns False."""
        doc = etree.parse(StringIO(self.rawrequest))
        if doc and doc.xpath('/g:GetFeature', namespaces={'g': NS_WFS}):
            return True
        return False

    def bill(self, db_conn=None):
        """ This method should be called to take account for the request. If
        this object represents a GetFeature request, a modified version of this
        request's raw data is used to query the number of feature that would
        have been retrieved by this GetFeature request. If the modified version
        of this request returns a valid quantity, this number is used to update
        the row in the database. If everything was ok in this process, this
        method returns the number of features of this request, otherwise
        False."""
        doc = etree.parse(StringIO(self.rawrequest))
        req = doc.xpath('/g:GetFeature', namespaces={'g': NS_WFS})

        if not req or not req[0]:
            logger.error("Could not find request parameters for request id"
                        "%r" % self.id)
            return False

        req[0].set('resultType', 'hits')
        opener = urllib.URLopener()
        data   = etree.tostring(doc)
        conn   = opener.open(self.wfsidintern, data)

        if not conn:
            logger.error("Could not connect to host '%r'." %
                         self.wfsidintern)
            return False

        response = etree.parse(conn)
        return self.handle_response(response)


def open_connection():
    host     = get_config_option("database", "host", "localhost")
    dbname   = get_config_option("database", "database")
    user     = get_config_option("database", "user")
    password = get_config_option("database", "password", "")

    try:
        conn = db.connect(host=host, database=dbname, user=user,
                          password=password)
        return conn
    except:
        logger.critical("Could not connect to database (host='%s', database="
                     "'%s', user='%s'" % (host, dbname, user))


def close_connection(*cur):
    for c in cur:
        if c:
            try: c.close()
            except: pass


def load_requests():
    requests = []
    try:
        conn = open_connection()
        if not conn:
            return None

        cur = conn.cursor()
        cur.execute(SQL_SELECT_OPEN_BILLINGS)
        while True:
            row = cur.fetchone()
            if not row: break

            req = None
            if row[3] == REQUEST_POST:
                req = POSTRequest(*row)
            elif row[3] == REQUEST_GET:
                req = GETRequest(*row)
            else:
                logger.warn("Unknown request format found: %r" % row[3])

            if req and not req.cashtime and req.is_get_feature_request():
                requests.append(req)
        close_connection(cur, conn)
        return requests
    except:
        logger.error("An error occured while connecting to database.")
        return None


def main():
    parser = OptionParser()
    parser.add_option("-c", "--config-file", default=DEFAULT_CONFIG)
    opts, rest = parser.parse_args()

    global config
    config=read_config(opts.config_file)

    setup_logging()
    print "Start billing script."
    print "===============================================\n"

    try:
        requests = load_requests()
        if not requests:
            return

        logger.info("Found %d open billings for GetFeature requests." %
                    len(requests))

        failed  = 0
        success = 0
        for req in requests:
            logger.info("================================")
            logger.info("Process request with db id %d (format = %d)" % (
                        req.id, req.requestformat))
            try:
                if not req.bill():
                    failed += 1
                else:
                    success += 1
            except:
                logger.error("Could not bill the request with id %d" % req.id)

        print "\nScript ended successfully."
        print "\n==============================================="
        print "== Summary: Found %d open requests." % len(requests)
        print "== Summary: Sucessfully billed %d requests." % success
        print "== Summary: %d requests failed." % failed
        print "== (See the logfile '%s' for more details)" % get_config_option(
                                                                "logging",
                                                                "logfile",
                                                                DEFAULT_LOGFILE)
    except:
        logger.error("An unexpected error occured - the script did not "
                     "finish regularly!")


if __name__ == '__main__':
    main()
