# Copyright (C) 2007, 2008, 2010 by Intevation GmbH
# Authors:
# Bernhard Herzog <bh@intevation.de>
#
# This program is free software under the GPL (>=v2)
# Read the file GPL.txt coming with the software for details.

"""Database backend of OSAAS"""


def oracle_sequence_nextval(name):
    """Returns the Oracle SQL expression fetching the next value of a sequence.
    """
    return "%s.nextval" % name

def postgresql_sequence_nextval(name):
    """Returns the POstGreSQL expression fetching the next value of a sequence.
    """
    return "nextval('%s')" % name

db_api_names = dict(
    postgresql=("psycopg2", dict(osaas_nextval=postgresql_sequence_nextval)),
    oracle=("cx_Oracle", dict(osaas_nextval=oracle_sequence_nextval))
)

def load_db_api(dbtype):
    """Loads the DB API module indicated by the dbtype parameter.  The
    dbtype is looked up in the db_api_names dict, which matches the
    dbtype to a concrete python module name, among other things.  The
    function returns the python module with the DB API.
    """
    modulename, extra_api = db_api_names.get(dbtype, (None, None))
    if modulename is None:
        raise ValueError("Unknown dbtype: %r; must be one of %s"
                         % (dbtype, ", ".join(db_api_names.keys())))
    module = __import__(modulename)
    for key, value in extra_api.items():
        setattr(module, key, value)

    return module


class DBField(object):

    def __init__(self, parameter, column, column_type):
        self.parameter = parameter
        self.column = column
        self.column_type = column_type


class Insertion(object):

    def __init__(self, table, fields):
        self.table = table
        self.fields = fields


class DBBackend(object):

    def __init__(self, db_api, connection_parameters, rules, logger=None):
        self.db_api = db_api
        self.connection = None
        params, user, password, dsn = connection_parameters
        if params:
            connection_parameters = (params,)
        else:
            connection_parameters = (user, password, dsn)
        self.connection_parameters = connection_parameters
        self.rules = rules
        self.logger = logger
        self.insert_statement = None

    def log_debug(self, message, *args):
        if self.logger is not None:
            self.logger.debug(message, *args)

    def log_error(self, message, *args):
        if self.logger is not None:
            self.logger.error(message, *args)

    def log_exception(self, message, *args):
        if self.logger is not None:
            self.logger.exception(message, *args)

    def ensure_connection(self, force_new_connection=False):
        if force_new_connection and self.connection is not None:
            try:
                self.connection.close()
            except:
                self.log_exception("Exception during connection.close()"
                                   " while forcing a new connection")
            self.connection = None
        if self.connection is None:
            self.log_debug("Connecting to database")
            self.connection = self.connect(*self.connection_parameters)

    def connect(self, *args):
        return self.db_api.connect(*args)

    def named_parameter(self, name):
        """Return a named parameter reference according to the db_api used"""
        paramstyle = self.db_api.paramstyle
        if paramstyle == "named":
            template = ":%s"
        elif paramstyle == "pyformat":
            template = "%%(%s)s"
        else:
            raise RuntimeError("Unsupported DB-API paramstyle: %r", paramstyle)

        return template % name

    def get_insert_statement(self):
        if self.insert_statement is None:
            rule = self.rules[0]
            names = ", ".join([f.column for f in rule.fields])
            statement = ("INSERT INTO %s (ID, %s) VALUES (%s, %s)"
                         % (rule.table, names,
                            self.db_api.osaas_nextval(rule.table + "_seq"),
                            ", ".join([self.named_parameter(f.column)
                                       for f in rule.fields])))
            self.insert_statement = statement
        return self.insert_statement

    def handle_record(self, record):
        """Write record into the database.  Log the record if unsuccessful."""
        try:
            self.insert_record_with_reconnect(record)
        except:
            self.log_error("could not write record %s", record.raw_record)
            raise

    def insert_record_with_reconnect(self, record):
        """Insert record into the database, reconnecting on some errors"""
        try:
            try:
                self.insert_record(record)
            except self.db_api.OperationalError:
                self.log_exception("Got OperationalError."
                                   " Trying again with a new connection"
                                   " in case it's because of a DB server"
                                   " connection problem")
                self.ensure_connection(force_new_connection=True)
                self.insert_record(record)
        except:
            # make sure to always use a new connection after an error
            # happened.  If the network connection to the database
            # vanishes unexpectedly, the DB API implementation might be
            # left in an unusable state.  Also, we cannot just do this
            # when an OperationalError is raised because some DB API
            # implementations may raise different exceptions and even
            # conforming implementations may raise other errors if the
            # implementation
            self.connection = None
            raise

    def insert_record(self, record):
        """Insert record into the database."""
        statement = self.get_insert_statement()
        values = {}
        for f in self.rules[0].fields:
            values[f.column] = getattr(record, f.parameter)

        self.ensure_connection()
        cursor = self.connection.cursor()
        try:
            cursor.execute(statement, values)
        except:
            self.connection.rollback()
            raise
        else:
            self.connection.commit()

    def create_tables(self):
        """Creates the tables and sequences for the rules"""
        self.ensure_connection()
        cursor = self.connection.cursor()
        try:
            for rule in self.rules:
                cursor.execute("CREATE SEQUENCE %s" % (rule.table + "_seq"))

                lines = []
                lines.append("CREATE TABLE %s" % rule.table)
                lines.append("(")
                fieldlines = ["    ID INT PRIMARY KEY"]
                for field in rule.fields:
                    fieldlines.append("    %s %s"
                                      % (field.column, field.column_type))
                lines.append(",\n".join(fieldlines))
                lines.append(")")
                statement = "\n".join(lines)
                cursor.execute(statement)
        except:
            self.connection.rollback()
            raise
        else:
            self.connection.commit()
