# Copyright (C) 2007, 2008, 2010 by Intevation GmbH
# Authors:
# Bernhard Herzog <bh@intevation.de>
#
# This program is free software under the GPL (>=v2)
# Read the file GPL.txt coming with the software for details.

"""Tests for osaas.server"""

import base64
import threading
import datetime

import serversupport
import support

from osaas.server import OSAASServer


class OSAASServerTests(serversupport.ServerTest, support.FileTestMixin):

    def setUp(self):
        self.userdbfile = self.create_temp_file(self.id() + "-userdb",
                        "karl:{SSHA}NGVfMfEuUjT3wFNgP0KiqWs47m9zYWx6")
        self.authentication_headers = [
            ("Authorization", "Basic %s" % base64.b64encode("karl:xyzzy")),
            ]
        serversupport.ServerTest.setUp(self)

    def create_server(self, *args, **kw):
        return OSAASServer(userdbfile=self.userdbfile, *args, **kw)

    def test_post_wrong_path(self):
        self.run_simple_http_test("/some/path", 404, method="POST",
                                  extra_headers=self.authentication_headers)

    def test_post_no_content_type(self):
        self.run_simple_http_test("/owsaccounting/", 415, method="POST",
                                  extra_headers=self.authentication_headers)

    def test_post_no_content_length(self):
        extra_headers = [("Content-Type", "application/x-www-form-urlencoded")]
        self.run_simple_http_test("/owsaccounting/", 411, method="POST",
                                  extra_headers=(extra_headers
                                                 + self.authentication_headers))

    def test_unauthenticated(self):
        body = "requeststring=SERVICE=WMS%26REQUEST=GetMap"
        extra_headers = [("Content-Type", "application/x-www-form-urlencoded"),
                         ("Content-Length", len(body))]
        self.run_simple_http_test("/owsaccounting/", 401, method="POST",
                                  extra_headers=extra_headers,
                                  body=body)


    def test_malformed_basic_authentication(self):
        # If the authorization were successful we would get 415 errors
        # because of the missing content type.  With malformed
        # authentication headers we should get 401.

        # sanity check first: do we get 415 with correct authentication?
        self.run_simple_http_test("/owsaccounting/", 415, method="POST",
                                  extra_headers=self.authentication_headers)

        malformed = [("Authorization",
                      "Basic %s" % base64.b64encode("hans 123"))]
        self.run_simple_http_test("/owsaccounting/", 401, method="POST",
                                  extra_headers=malformed)

    def test_malformed_basic_authentication2(self):
        # If the authorization were successful we would get 415 errors
        # because of the missing content type.  With malformed
        # authentication headers we should get 401.

        # sanity check first: do we get 415 with correct authentication?
        self.run_simple_http_test("/owsaccounting/", 415, method="POST",
                                  extra_headers=self.authentication_headers)

        malformed = [("Authorization", "Basic ")]
        self.run_simple_http_test("/owsaccounting/", 401, method="POST",
                                  extra_headers=malformed)

    def test_authentication_header_but_not_basic(self):
        # If the authorization were successful we would get 415 errors
        # because of the missing content type.  With malformed
        # authentication headers we should get 401.

        # sanity check first: do we get 415 with correct authentication?
        self.run_simple_http_test("/owsaccounting/", 415, method="POST",
                                  extra_headers=self.authentication_headers)

        malformed = [("Authorization", "Invented data")]
        self.run_simple_http_test("/owsaccounting/", 401, method="POST",
                                  extra_headers=malformed)


class OSAASDBServerTests(serversupport.ServerTest, support.AttributeTestMixin,
                         support.FileTestMixin):

    def setUp(self):
        self.dbrequests = []
        self.dbrequest_event = threading.Event()
        self.userdbfile = self.create_temp_file(self.id() + "-userdb",
                        "karl:{SSHA}NGVfMfEuUjT3wFNgP0KiqWs47m9zYWx6")
        self.authentication_headers = [
            ("Authorization", "Basic %s" % base64.b64encode("karl:xyzzy")),
            ]
        serversupport.ServerTest.setUp(self)

    def handle_record(self, req):
        self.dbrequests.append(req)
        self.dbrequest_event.set()

    def create_server(self, *args, **kw):
        return OSAASServer(dbbackend=self, userdbfile=self.userdbfile,
                           *args, **kw)

    def send_formdata_request(self, formdata, expected_status=200):
        extra_headers = ([("Content-Type", "application/x-www-form-urlencoded"),
                         ("Content-Length", len(formdata))]
                         + self.authentication_headers)
        return self.run_simple_http_test("/owsaccounting/", expected_status,
                                         method="POST",
                                         extra_headers=extra_headers,
                                         body=formdata)

    def test_post_with_content(self):
        formdata = "&".join([
            "user=karl",
            "starttime=2007-06-01T12:14:10.000Z",
            "endtime=2007-06-01T12:14:16.321Z",
            "wmsidextern=example.com%2Fcgi-bin%2Ffridawms",
            "wmsidintern=localhost%2Fcgi-bin%2Fmyfrida",
            "requeststring=VERSION=1.1.1%26REQUEST=GetMap"
            "%26FORMAT=image%2Fpng%26TRANSPARENT=TRUE"
            "%26WIDTH=460%26HEIGHT=348"
            "%26EXCEPTIONS=application%2Fvnd.ogc.se_xml"
            "%26BGCOLOR=0xffffff%26BBOX=0.0,0.0,460.0,348.0"
            "%26LAYERS=gewaesser%26STYLES=%26SRS=EPSG:4326"
            "%26SERVICE=WMS"])
        self.send_formdata_request(formdata)
        self.dbrequest_event.wait(1)
        self.assertEquals(len(self.dbrequests), 1)
        self.check_attributes(self.dbrequests[0],
                              user="karl",
                              starttime=datetime.datetime(2007, 6, 1,
                                                          12, 14, 10, 0),
                              endtime=datetime.datetime(2007, 6, 1,
                                                        12, 14, 16, 321000),
                              wmsidextern="example.com/cgi-bin/fridawms",
                              wmsidintern="localhost/cgi-bin/myfrida",
                              requeststring=("VERSION=1.1.1&REQUEST=GetMap"
                                             "&FORMAT=image/png"
                                             "&TRANSPARENT=TRUE"
                                             "&WIDTH=460&HEIGHT=348"
                                      "&EXCEPTIONS=application/vnd.ogc.se_xml"
                                             "&BGCOLOR=0xffffff"
                                             "&BBOX=0.0,0.0,460.0,348.0"
                                             "&LAYERS=gewaesser&STYLES="
                                             "&SRS=EPSG:4326"
                                             "&SERVICE=WMS"),
                              VERSION="1.1.1",
                              REQUEST="GetMap",
                              FORMAT="image/png",
                              TRANSPARENT="TRUE",
                              WIDTH="460",
                              HEIGHT="348",
                              EXCEPTIONS="application/vnd.ogc.se_xml",
                              BGCOLOR="0xffffff",
                              BBOX="0.0,0.0,460.0,348.0",
                              LAYERS="gewaesser",
                              STYLES="",
                              SRS="EPSG:4326",
                              SERVICE="WMS")


    def test_post_with_content_and_filtering(self):
        self.send_formdata_request("&".join([
            "user=karl",
            "starttime=2007-06-01T12:14:09.123Z",
            "endtime=2007-06-01T12:14:16.515Z",
            "wmsidextern=example.com%2Fcgi-bin%2Ffridawms",
            "wmsidintern=localhost%2Fcgi-bin%2Fmyfrida",
            "requeststring=VERSION=1.1.1%26REQUEST=GetCapabilities"
                           "%26SERVICE=WMS"]))
        self.send_formdata_request("&".join([
            "user=karl",
            "starttime=2007-06-01T12:14:17.832Z",
            "endtime=2007-06-01T12:15:20.746Z",
            "wmsidextern=example.com%2Fcgi-bin%2Ffridawms",
            "wmsidintern=localhost%2Fcgi-bin%2Fmyfrida",
            "requeststring=VERSION=1.1.1%26REQUEST=GetMap"
            "%26FORMAT=image%2Fpng%26TRANSPARENT=TRUE"
            "%26WIDTH=460%26HEIGHT=348"
            "%26EXCEPTIONS=application%2Fvnd.ogc.se_xml"
            "%26BGCOLOR=0xffffff%26BBOX=0.0,0.0,460.0,348.0"
            "%26LAYERS=gewaesser%26STYLES=%26SRS=EPSG:4326"
            "%26SERVICE=WMS"]))
        self.dbrequest_event.wait(1)
        self.assertEquals(len(self.dbrequests), 1)
        self.check_attributes(self.dbrequests[0],
                              user="karl",
                              starttime=datetime.datetime(2007, 6, 1,
                                                          12, 14, 17, 832000),
                              endtime=datetime.datetime(2007, 6, 1,
                                                        12, 15, 20, 746000),
                              wmsidextern="example.com/cgi-bin/fridawms",
                              wmsidintern="localhost/cgi-bin/myfrida",
                              requeststring=("VERSION=1.1.1&REQUEST=GetMap"
                                             "&FORMAT=image/png"
                                             "&TRANSPARENT=TRUE"
                                             "&WIDTH=460&HEIGHT=348"
                                      "&EXCEPTIONS=application/vnd.ogc.se_xml"
                                             "&BGCOLOR=0xffffff"
                                             "&BBOX=0.0,0.0,460.0,348.0"
                                             "&LAYERS=gewaesser&STYLES="
                                             "&SRS=EPSG:4326"
                                             "&SERVICE=WMS"),
                              VERSION="1.1.1",
                              REQUEST="GetMap",
                              FORMAT="image/png",
                              TRANSPARENT="TRUE",
                              WIDTH="460",
                              HEIGHT="348",
                              EXCEPTIONS="application/vnd.ogc.se_xml",
                              BGCOLOR="0xffffff",
                              BBOX="0.0,0.0,460.0,348.0",
                              LAYERS="gewaesser",
                              STYLES="",
                              SRS="EPSG:4326",
                              SERVICE="WMS")

    def test_post_without_service_field(self):
        self.send_formdata_request("&".join([
            "user=karl", "starttime=2007-06-01T12:14:16.010Z",
            "endtime=2007-06-01T12:14:17.987Z",
            "wmsidextern=example.com%2Fcgi-bin%2Ffridawms",
            "wmsidintern=localhost%2Fcgi-bin%2Fmyfrida",
            "requeststring=VERSION=1.1.1%26REQUEST=GetCapabilities"]))
        self.send_formdata_request("&".join([
            "user=karl",
            "starttime=2007-06-01T12:14:18.020Z",
            "endtime=2007-06-01T12:14:22.000Z",
            "wmsidextern=example.com%2Fcgi-bin%2Ffridawms",
            "wmsidintern=localhost%2Fcgi-bin%2Fmyfrida",
            "requeststring=VERSION=1.1.1%26REQUEST=GetMap"
            "%26FORMAT=image%2Fpng%26TRANSPARENT=TRUE"
            "%26WIDTH=460%26HEIGHT=348"
            "%26EXCEPTIONS=application%2Fvnd.ogc.se_xml"
            "%26BGCOLOR=0xffffff%26BBOX=0.0,0.0,460.0,348.0"
            "%26LAYERS=gewaesser%26STYLES=%26SRS=EPSG:4326"]))
        self.dbrequest_event.wait(1)
        self.assertEquals(len(self.dbrequests), 1)
        self.check_attributes(self.dbrequests[0],
                              user="karl",
                              starttime=datetime.datetime(2007, 6, 1,
                                                          12, 14, 18, 20000),
                              endtime=datetime.datetime(2007, 6, 1,
                                                        12, 14, 22, 0),
                              wmsidextern="example.com/cgi-bin/fridawms",
                              wmsidintern="localhost/cgi-bin/myfrida",
                              requeststring=("VERSION=1.1.1&REQUEST=GetMap"
                                             "&FORMAT=image/png"
                                             "&TRANSPARENT=TRUE"
                                             "&WIDTH=460&HEIGHT=348"
                                      "&EXCEPTIONS=application/vnd.ogc.se_xml"
                                             "&BGCOLOR=0xffffff"
                                             "&BBOX=0.0,0.0,460.0,348.0"
                                             "&LAYERS=gewaesser&STYLES="
                                             "&SRS=EPSG:4326"),
                              VERSION="1.1.1",
                              REQUEST="GetMap",
                              FORMAT="image/png",
                              TRANSPARENT="TRUE",
                              WIDTH="460",
                              HEIGHT="348",
                              EXCEPTIONS="application/vnd.ogc.se_xml",
                              BGCOLOR="0xffffff",
                              BBOX="0.0,0.0,460.0,348.0",
                              LAYERS="gewaesser",
                              STYLES="",
                              SRS="EPSG:4326")

    def test_post_with_content_with_missing_fields(self):
        # the formdata is missing the "endtime" field
        formdata = "&".join([
            "user=karl",
            "starttime=2007-06-01T12:14:18.020Z",
            "wmsidextern=example.com%2Fcgi-bin%2Ffridawms",
            "wmsidintern=localhost%2Fcgi-bin%2Fmyfrida",
            "requeststring=VERSION=1.1.1%26REQUEST=GetMap"
            "%26FORMAT=image%2Fpng%26TRANSPARENT=TRUE"
            "%26WIDTH=460%26HEIGHT=348"
            "%26EXCEPTIONS=application%2Fvnd.ogc.se_xml"
            "%26BGCOLOR=0xffffff%26BBOX=0.0,0.0,460.0,348.0"
            "%26LAYERS=gewaesser%26STYLES=%26SRS=EPSG:4326"
            "%26SERVICE=WMS"])
        response_text = self.send_formdata_request(formdata,
                                                   expected_status=400)
        self.failUnless("formdata is missing required field 'endtime'"
                        in response_text)
