# Copyright (C) 2007 by Intevation GmbH
# Authors:
# Bernhard Herzog <bh@intevation.de>
#
# This program is free software under the GPL (>=v2)
# Read the file GPL.txt coming with the software for details.

"""Tests for osaas.http.threadedserver"""

import httplib
from urlparse import urlparse
from threading import Event

from osaas.http.httpserver import HTTPRequestHandler
from osaas.http.threadedserver import ThreadedHTTPServer

from serversupport import ServerTest

class BlockingRequestHandler(HTTPRequestHandler):

    """RequestHandler that blocks on event object provided by the test case.
    See the actual test methods for documentation on how the tests and
    the request handler interact
    """

    def do_GET(self):
        try:
            events = self.server.test_case.events[self.path]
        except KeyError:
            self.send_error(404)
            return

        events[0].wait()

        try:
            self.send_response(200)
            self.end_headers()
            self.wfile.write(self.path)
        finally:
            # set events[1] in a finally block to make user it is always
            # set.  Also, it must be set after the response has been
            # sent.  Otherwise it can happen that the log entry is not
            # written until the another thread has written a log entry
            # which can make the test cases fail
            events[1].set()


class SimpleHTTPServerTests(ServerTest):

    def setUp(self):
        self.workerpool_log_stream = self.new_logger(self.error_logger_name
                                                     + ".workerpool")
        ServerTest.setUp(self)

    def create_server(self, *args, **kw):
        return ThreadedHTTPServer(RequestHandlerClass=BlockingRequestHandler,
                                  *args, **kw)

    def test_get_multiple(self):
        """Test that multiple concurrent requests can be handled out of order"""
        # The test basically sends two requests to the server and tests
        # whether they have been processed in reverse order.  For this
        # to work the test has to cooperate with the request handlers to
        # make sure the first request is accepted by the serve but not
        # actually processed until the second request has been
        # processed.  This is made more complex by the desire to make
        # sure that the test runs to completion even if the server
        # implementation can only process requests in order of arrival.
        #
        # Even if the server only processes requests in order, the
        # server and therefore the handlers run in their own thread but
        # in the same process as the test case, so we can use lock based
        # thread synchronization mechanisms.  In this case we use Event
        # objects stored in the test case's instance variable events.
        # The handlers can access the test case object through the serve
        # object (see setUp).
        #
        # The events attribute is a dictionary mapping the paths used in
        # HTTP GET requests to pairs of Event objects.  The test case
        # creates two such pairs and sends corresponding GET requests to
        # the server.  The handlers block on the first event object of
        # the relevant pair.  The test case signals the first event of
        # the request sent last and then waits for the second event with
        # a timeout.  The requests will signal the second event after
        # the first event has been signaled.  Normally, if the events
        # are processed in parallel, the request startet second will
        # already be waiting for the event and immediatly signal the
        # other event.  If the requests are processed in order, the test
        # case will wait until timeout.  In both cases the test case
        # then signals the event of the first request.  Afterwards the
        # responses of both requests are read and the log file is
        # checked whether the requests were actually processed in the
        # expected order.

        self.events = {}
        eventnames = ["/event1", "/event2"]

        connections = {}
        for name in eventnames:
            self.events[name] = (Event(), Event())
            http = httplib.HTTPConnection("localhost", self.server.server_port)
            http.request("GET", name)
            connections[name] = http

        # unblock the second event first and then the first event and
        # wait for it to report back.  The unblock the first event.
        self.events[eventnames[1]][0].set()
        self.events[eventnames[1]][1].wait(1.0)
        self.events[eventnames[0]][0].set()

        for name in eventnames:
            http = connections[name]
            response = http.getresponse()
            self.assertEquals(response.status, 200)
            self.assertEquals(response.read(), name)

        self.check_log(['localhost - - [] "GET /event2 HTTP/1.1" 200 -',
                        'localhost - - [] "GET /event1 HTTP/1.1" 200 -'])
        self.assertEquals(self.workerpool_log_stream.getvalue(),
                          "Starting 5 threads\n")
